<?php

namespace AltSolution\Admin\Form;

use Illuminate\Support\Arr;
use IteratorAggregate;
use Illuminate\Contracts\Validation\Factory;
use Illuminate\Http\Request;
use Illuminate\Support\Str;

class Form implements IteratorAggregate
{
    const LAYOUT_VERTICAL = 'vertical';
    const LAYOUT_HORIZONTAL = 'horizontal';
    const LAYOUT_INLINE = 'inline';

    /** @var ComponentInterface[] */
    private $components = [];

    /** @var  mixed */
    private $dataSource;

    /**
     * Available vertical, horizontal, inline
     * @var string
     */
    private $layout;

    /**
     * Add component to form
     * @param ComponentInterface $component
     */
    public function add(ComponentInterface $component)
    {
        $name = $component->getName();
        $this->components[$name] = $component;
        $component->setForm($this);
    }

    /**
     * Get component by name
     * @param string $name
     * @return ComponentInterface
     */
    public function get($name)
    {
        return $this->components[$name];
    }

    /**
     * Check if form has component by name
     * @param string $name
     * @return bool
     */
    public function has($name)
    {
        return isset($this->components[$name]);
    }

    /**
     * Render form component
     * @param string $name
     * @param array $options
     * @return string
     */
    public function renderComponent($name, array $options = [])
    {
        $component = $this->components[$name];

        return $component->render($options);
    }

    /**
     * Set form field data source
     * @param mixed $dataSource
     */
    public function setDataSource($dataSource)
    {
        $this->dataSource = $dataSource;
    }

    /**
     * Get any value from data source
     * @param string|null $key
     * @return mixed
     */
    public function getData($key = null)
    {
        if ($key === null) {
            return $this->dataSource;
        }
        return Arr::get($this->dataSource, $key);
    }

    /**
     * @return array
     */
    public function extractData()
    {
        $result = [];
        foreach ($this->getFields() as $field) {
            $fieldName = $field->getName();
            Arr::set($result, $fieldName, $this->getData($fieldName));
        }
        return $result;
    }

//    /**
//     * Call function of data source
//     * @param string $fn
//     * @param array $args
//     * @return mixed
//     */
//    public function callData($fn, array $args = [])
//    {
//        return call_user_func_array([$this->dataSource, $fn], $args);
//    }

    public function setLayout(string $layout): void
    {
        $available = [
            self::LAYOUT_VERTICAL,
            self::LAYOUT_HORIZONTAL,
            self::LAYOUT_INLINE
        ];
        if (!in_array($layout, $available, true)) {
            throw new \InvalidArgumentException("Invalid layout '{$layout}'");
        }

        $this->layout = $layout;
    }

    public function getLayout(): string
    {
        if ($this->layout === null) {
            return config('admin.form_layout');
        }
        return $this->layout;
    }

    // --- IteratorAggregate

    public function getIterator()
    {
        $iter = new \ArrayIterator($this->components);
        $filter = new \CallbackFilterIterator($iter, function(ComponentInterface $component) {
            return $component->isEnumerable();
        });
        return $filter;
    }

    /**
     * @return \Generator|FieldInterface[]
     */
    public function getFields()
    {
        foreach ($this->components as $component) {
            if ($component instanceof FieldInterface) {
                yield $component;
            }
        }
    }

    /**
     * @param Request $request
     * @param array $customRules
     * @param array $customMessages
     * @param array $customAttributes
     * @return array
     * @see \Illuminate\Foundation\Validation\ValidatesRequests
     */
    public function validate(Request $request, array $customRules = [], array $customMessages = [], array $customAttributes = [])
    {
        $rules = $this->getValidationRules();
        // merge rules
        foreach ($customRules as $ruleAttr => $rule) {
            if (is_array($rule)) {
                $ruleArr = $rule;
            } elseif (is_string($rule)) {
                $ruleArr = explode('|', $rule);
            } else {
                $ruleArr = [$rule];
            }
            if (isset($rules[$ruleAttr])) {
                $rules[$ruleAttr] = array_merge($rules[$ruleAttr], $ruleArr);
            } else {
                $rules[$ruleAttr] = $ruleArr;
            }
        }

        $formMessages = [];
        $messages = array_merge($formMessages, $customMessages);

        $formAttributes = $this->getValidationAttributes();
        $attributes = array_merge($formAttributes, $customAttributes);

        app(Factory::class)
            ->make($request->all(), $rules, $messages, $attributes)
            ->validate();

        return $request->only(collect($rules)->keys()->map(function ($rule) {
            return Str::contains($rule, '.') ? explode('.', $rule)[0] : $rule;
        })->unique()->toArray());
    }

    /**
     * @return array
     */
    public function getValidationRules()
    {
        $rules = [];
        foreach ($this->getFields() as $field) {
            // если поле не доступно на редактирование пользователю,
            // то не валидировать и в следствии не обновлять его в модели
            if ($field->isFillable()) {
                $rules[$field->getValidationName()] = $field->getValidationRules();
            }
        }

        return $rules;
    }

    public function getValidationAttributes()
    {
        $attributes = [];
        foreach ($this->getFields() as $field) {
            $attributes[$field->getValidationName()] = $field->getValidationAttribute();
        }

        return $attributes;
    }
}